@@ -4,9 +4,11 @@ import os, sys
 import warnings
 import logging as logg
 from enum import Enum
-from collections import Mapping, Sequence, Sized, OrderedDict
+from collections import OrderedDict
 from functools import reduce
-from typing import Union
+from pathlib import Path
+from textwrap import indent, dedent
+from typing import Union, Optional, Any, Iterable, Mapping, Sequence, Sized, Tuple
 from copy import deepcopy
 
 import numpy as np
@@ -17,7 +19,6 @@ from pandas.api.types import is_string_dtype, is_categorical
 from scipy import sparse
 from scipy.sparse import issparse
 from scipy.sparse.sputils import IndexMixin
-from textwrap import dedent
 from natsort import natsorted
 
 from . import h5py
@@ -28,22 +29,19 @@ logg.basicConfig(format=FORMAT, level=logg.INFO, stream=sys.stdout)
 
 _MAIN_NARRATIVE = """\
 :class:`~anndata.AnnData` stores a data matrix ``.X`` together with annotations
-of observations ``.obs``, variables ``.var`` and unstructured annotations
-``.uns``.
+of observations ``.obs``, variables ``.var`` and unstructured annotations ``.uns``.
 
-.. raw:: html
-
-    <img src="http://falexwolf.de/img/scanpy/anndata.svg"
-         style="width: 350px; margin: 10px 0px 15px 20px">
+.. figure:: https://falexwolf.de/img/scanpy/anndata.svg
+   :width: 350px
 
 An :class:`~anndata.AnnData` object ``adata`` can be sliced like a pandas
-dataframe, for instance, ``adata_subset = adata[:,
-list_of_variable_names]``. :class:`~anndata.AnnData`'s basic structure is
-similar to R's ExpressionSet [Huber15]_. If setting an `.h5ad`-formatted HDF5
-backing file ``.filename``, data remains on the disk but is automatically loaded
-into memory if needed. See this `blog post
-<http://falexwolf.de/blog/171223_AnnData_indexing_views_HDF5-backing/>`_ for
-more details.
+dataframe, for instance, ``adata_subset = adata[:, list_of_variable_names]``.
+:class:`~anndata.AnnData`'s basic structure is similar to R's ExpressionSet
+[Huber15]_. If setting an ``.h5ad``-formatted HDF5 backing file ``.filename``,
+data remains on the disk but is automatically loaded into memory if needed.
+See this `blog post`_ for more details.
+
+.. _blog post: http://falexwolf.de/blog/171223_AnnData_indexing_views_HDF5-backing/
 """
 
 
@@ -64,15 +62,15 @@ class BoundRecArr(np.recarray):
     """
     _attr_choices = ['obsm', 'varm']
 
-    def __new__(cls, input_array, parent, attr):
+    def __new__(cls, input_array: np.ndarray, parent: Any, attr: str):
         """
         Parameters
         ----------
-        source : `np.ndarray`
+        input_array
             A (structured) numpy array.
-        parent : object
+        parent
             Any object to which the BoundRecArr shall be bound to.
-        attr : string
+        attr
             The name of the attribute as which it appears in parent.
         """
         arr = np.asarray(input_array).view(cls)
@@ -191,7 +189,7 @@ def df_to_records_fixed_width(df):
         dtype={'names': names, 'formats': formats}), uns
 
 
-def _fix_shapes(X, single_col=False):
+def _fix_shapes(X):
     """Fix shapes of array or sparse matrix.
 
     Flatten 2d array with a single row or column to emulate numpys
@@ -210,9 +208,6 @@ def _fix_shapes(X, single_col=False):
     elif len(X.shape) == 1:
         n_vars = X.shape[0]
         n_obs = 1
-    elif len(X.shape) == 1 and single_col:
-        n_obs = X.shape[0]
-        n_vars = 1
     else:
         n_vars = 1
         n_obs = 1
@@ -220,15 +215,16 @@ def _fix_shapes(X, single_col=False):
 
 
 def _normalize_index(index, names):
+    # the following is insanely slow for sequences, we replaced it using pandas below
     def name_idx(i):
         if isinstance(i, str):
             # `where` returns an 1-tuple (1D array) of found indices
-            i = np.where(names == i)[0]
-            if len(i) == 0:  # returns array of length 0 if nothing is found
+            i_found = np.where(names == i)[0]
+            if len(i_found) == 0:  # returns array of length 0 if nothing is found
                 raise IndexError(
-                    'Name "{}" is not valid observation/variable name.'
-                    .format(index))
-            i = i[0]
+                    'Key "{}" is not valid observation/variable name/index.'
+                    .format(i))
+            i = i_found[0]
         return i
 
     if isinstance(index, slice):
@@ -241,14 +237,24 @@ def _normalize_index(index, names):
         return slice(start, stop, step)
     elif isinstance(index, (int, str)):
         return name_idx(index)
-    elif isinstance(index, (Sequence, np.ndarray)):
-        return np.fromiter(map(name_idx, index), 'int64')
+    elif isinstance(index, (Sequence, np.ndarray, pd.Index)):
+        # here, we replaced the implementation based on name_idx with this
+        # incredibly faster one
+        positions = pd.Series(index=names, data=range(len(names)))
+        positions = positions[index]
+        if positions.isnull().values.any():
+            raise KeyError(
+                'Indices "{}" contain invalid observation/variables names/indices.'
+                .format(index))
+        return positions.values
     else:
         raise IndexError('Unknown index {!r} of type {}'
                          .format(index, type(index)))
 
 
 def _gen_dataframe(anno, length, index_names):
+    if isinstance(anno, pd.DataFrame):
+        return anno
     if anno is None or len(anno) == 0:
         _anno = pd.DataFrame(index=RangeIndex(0, length, name=None).astype(str))
     else:
@@ -278,7 +284,7 @@ class AnnDataFileManager:
         if self._filename is None:
             return 'Backing file manager: no file is set.'
         else:
-            return 'Backing file manager of file {}.'.format(self._filename)
+            return f'Backing file manager of file {self._filename}.'
 
     def __getitem__(self, key):
         return self._file[key]
@@ -516,29 +522,34 @@ class IndexDimError(IndexError):
         super().__init__(msg)
 
 
-class AnnData(IndexMixin):
-    __doc__ = dedent("""\
-    An annotated data matrix.
+class AnnData(IndexMixin, metaclass=utils.DeprecationMixinMeta):
+    __doc__ = dedent(f"""An annotated data matrix.
 
-    {main_narrative}
+    {indent(_MAIN_NARRATIVE, 4*' ')}
 
     Parameters
     ----------
-    X : `np.ndarray`, `sparse.spmatrix`
+    X
         A #observations × #variables data matrix. A view of the data is used if the
         data type matches, otherwise, a copy is made.
-    obs : `pd.DataFrame`, `dict`, structured `np.ndarray` or `None`, optional (default: `None`)
+    obs
         Key-indexed one-dimensional observation annotation of length #observations.
-    var : `pd.DataFrame`, `dict`, structured `np.ndarray` or `None`, optional (default: `None`)
+    var
         Key-indexed one-dimensional variable annotation of length #variables.
-    uns : `dict` or `None`, optional (default: `None`)
+    uns
         Unstructured annotation for the whole dataset.
-    obsm : structured `np.ndarray`, optional (default: `None`)
+    obsm
         Key-indexed multi-dimensional observation annotation of length #observations.
-    varm : structured `np.ndarray`, optional (default: `None`)
+    varm
         Key-indexed multi-dimensional observation annotation of length #observations.
-    dtype : simple `np.dtype`, optional (default: 'float32')
+    dtype
         Data type used for storage.
+    shape
+        Shape tuple (#observations, #variables). Can only be provided if ``X`` is ``None``.
+    filename
+        Name of backing file. See :class:`anndata.h5py.File`.
+    filemode
+        Open mode of backing file. See :class:`anndata.h5py.File`.
 
     See Also
     --------
@@ -555,51 +566,24 @@ class AnnData(IndexMixin):
     -----
     Multi-dimensional annotations are stored in ``.obsm`` and ``.varm``.
 
-    If the unstructured annotations `.uns` contain a sparse matrix of shape
-    `.n_obs` × `.n_obs`, these are also sliced.
+    If the unstructured annotations ``.uns`` contain a sparse matrix of shape
+    ``.n_obs`` × ``.n_obs``, these are also sliced.
 
     :class:`~anndata.AnnData` stores observations (samples) of variables
     (features) in the rows of a matrix. This is the convention of the modern
     classics of statistics [Hastie09]_ and machine learning [Murphy12]_, the
     convention of dataframes both in R and Python and the established statistics
-    and machine learning packages in Python (`statsmodels
-    <http://www.statsmodels.org/stable/index.html>`_, `scikit-learn
-    <http://scikit-learn.org/>`_).
+    and machine learning packages in Python (statsmodels_, scikit-learn_).
 
-    A data matrix is flattened if either #observations (`n_obs`) or #variables
-    (`n_vars`) is 1, so that numpy's slicing behavior is reproduced::
+    A data matrix is flattened if either #observations (``n_obs``) or #variables
+    (``n_vars``) is 1, so that numpy's slicing behavior is reproduced::
 
         adata = AnnData(np.ones((2, 2)))
         adata[:, 0].X == adata.X[:, 0]
 
-    Attributes
-    ----------
-    X
-    filename
-    isbacked
-    isview
-    n_obs
-    n_vars
-    shape
-    obs
-    obsm
-    obs_names
-    raw
-    var
-    varm
-    var_names
-
-    Methods
-    -------
-    concatenate
-    copy
-    transpose
-    obs_names_make_unique
-    var_names_make_unique
-    write
-    write_csvs
-    write_loom
-    """).format(main_narrative=_MAIN_NARRATIVE)
+    .. _statsmodels: http://www.statsmodels.org/stable/index.html
+    .. _scikit-learn: http://scikit-learn.org/
+    """)
 
     _BACKED_ATTRS = ['X', 'raw.X']
 
@@ -619,11 +603,20 @@ class AnnData(IndexMixin):
     }
 
     def __init__(
-            self, X=None, obs=None, var=None, uns=None,
-            obsm=None, varm=None, raw=None,
-            dtype='float32', single_col=False,
-            filename=None, filemode=None,
-            asview=False, oidx=None, vidx=None):
+        self,
+        X: Optional[Union[np.ndarray, sparse.spmatrix]] = None,
+        obs: Optional[Union[pd.DataFrame, Mapping[Any, Iterable[Any]], np.ndarray]] = None,
+        var: Optional[Union[pd.DataFrame, Mapping[Any, Iterable[Any]], np.ndarray]] = None,
+        uns: Optional[Mapping[Any, Any]] = None,
+        obsm: Optional[Union[np.ndarray, Mapping[str, Sequence[Any]]]] = None,
+        varm: Optional[Union[np.ndarray, Mapping[str, Sequence[Any]]]] = None,
+        raw: Optional[Union[Raw, Mapping[str, Any]]] = None,
+        dtype: Union[np.dtype, str] = 'float32',
+        shape: tuple = None,
+        filename: Optional[Union[Path, str]] = None,
+        filemode: Optional[str] = None,
+        asview: bool = False,
+        *, oidx=None, vidx=None):
         if asview:
             if not isinstance(X, AnnData):
                 raise ValueError('`X` has to be an AnnData object.')
@@ -632,7 +625,7 @@ class AnnData(IndexMixin):
             self._init_as_actual(
                 X=X, obs=obs, var=var, uns=uns,
                 obsm=obsm, varm=varm, raw=raw,
-                dtype=dtype, single_col=single_col,
+                dtype=dtype, shape=shape,
                 filename=filename, filemode=filemode)
 
     def _init_as_view(self, adata_ref, oidx, vidx):
@@ -690,6 +683,9 @@ class AnnData(IndexMixin):
             self._raw = None
 
     def _init_X_as_view(self):
+        if self._adata_ref.X is None:
+            self._X = None
+            return
         X = self._adata_ref.X[self._oidx, self._vidx]
         if len(X.shape) == 2:
             n_obs, n_vars = X.shape
@@ -710,7 +706,7 @@ class AnnData(IndexMixin):
     def _init_as_actual(
             self, X=None, obs=None, var=None, uns=None,
             obsm=None, varm=None, raw=None,
-            dtype='float32', single_col=False,
+            dtype='float32', shape=None,
             filename=None, filemode=None):
         from .readwrite.read import _read_h5ad
 
@@ -765,6 +761,8 @@ class AnnData(IndexMixin):
                 class_names = ', '.join(c.__name__ for c in StorageType.classes())
                 raise ValueError('`X` needs to be of one of {}, not {}.'
                                  .format(class_names, type(X)))
+            if shape is not None:
+                raise ValueError('`shape` needs to be `None` is `X` is not `None`.')
             # if type doesn't match, a copy is made, otherwise, use a view
             if issparse(X) or isinstance(X, ma.MaskedArray):
                 # TODO: maybe use view on data attribute of sparse matrix
@@ -776,8 +774,20 @@ class AnnData(IndexMixin):
             self._X, self._n_obs, self._n_vars = _fix_shapes(X)
         else:
             self._X = None
-            self._n_obs = len(obs)
-            self._n_vars = len(var)
+            self._n_obs = len([] if obs is None else obs)
+            self._n_vars = len([] if var is None else var)
+            # check consistency with shape
+            if shape is not None:
+                if self._n_obs == 0:
+                    self._n_obs = shape[0]
+                else:
+                    if self._n_obs != shape[0]:
+                        raise ValueError('`shape` is inconsistent with `obs`')
+                if self._n_vars == 0:
+                    self._n_vars = shape[1]
+                else:
+                    if self._n_vars != shape[1]:
+                        raise ValueError('`shape` is inconsistent with `var`')
 
         # annotations
         self._obs = _gen_dataframe(obs, self._n_obs,
@@ -817,7 +827,10 @@ class AnnData(IndexMixin):
             else:
                 # is dictionary from reading the file
                 self._raw = Raw(
-                    self, X=raw['X'], var=raw['var'], varm=raw['varm'])
+                    self,
+                    X=raw['X'],
+                    var=_gen_dataframe(raw['var'], raw['X'].shape[1], ['var_names', 'col_names']),
+                    varm=raw['varm'] if 'varm' in raw else None)
 
         # clean up old formats
         self._clean_up_old_format(uns)
@@ -855,8 +868,8 @@ class AnnData(IndexMixin):
         return self.n_obs, self.n_vars
 
     @property
-    def X(self) -> Union[np.ndarray, sparse.spmatrix]:
-        """Data matrix of shape `n_obs` × `n_vars` (`np.ndarray`, `sp.sparse.spmatrix`)."""
+    def X(self) -> Union[np.ndarray, sparse.spmatrix, None]:
+        """Data matrix of shape `n_obs` × `n_vars` (`np.ndarray`, `sp.sparse.spmatrix`) or `None`."""
         if self.isbacked:
             if not self.file.isopen: self.file.open()
             X = self.file['X']
@@ -866,6 +879,13 @@ class AnnData(IndexMixin):
 
     @X.setter
     def X(self, value):
+        if value is None:
+            if self.isview:
+                raise ValueError('Copy the view before setting the data matrix to `None`.')
+            if self.isbacked:
+                raise ValueError('Not implemented.')
+            self._X = None
+            return
         var_get = self.n_vars == 1 and self.n_obs == len(value)
         obs_get = self.n_obs == 1 and self.n_vars == len(value)
         if var_get or obs_get or self.shape == value.shape:
@@ -878,7 +898,7 @@ class AnnData(IndexMixin):
                 if self.isview:
                     # exit the view if we go from sparse to dense
                     if (issparse(value) and not issparse(self._adata_ref._X)
-                            or not issparse(value) and issparse(self._adata_ref._X)):
+                        or not issparse(value) and issparse(self._adata_ref._X)):
                         self._init_as_actual(self.copy())
                         self._X = value
                     else:
@@ -892,11 +912,15 @@ class AnnData(IndexMixin):
 
     @property
     def raw(self):
-        """Store raw version of `.X` and `.var` as `.raw.X` and `.raw.var`.
+        """Store raw version of ``.X`` and ``.var`` as ``.raw.X`` and ``.raw.var``.
 
-        The `.raw` attribute is initialized with the current content of an object by setting: ``adata.raw = adata``.
+        The `.raw` attribute is initialized with the current content of an object by setting::
 
-        Its content can be deleted by setting it back to `None`:  ``adata.raw = None``.
+            adata.raw = adata
+
+        Its content can be deleted by setting it back to ``None``::
+
+            adata.raw = None
         """
         return self._raw
 
@@ -957,6 +981,8 @@ class AnnData(IndexMixin):
 
     @uns.setter
     def uns(self, value):
+        if not isinstance(value, Mapping):
+            raise ValueError('Only dicitionary types are allowed for `.uns`.')
         if self.isview:
             # here, we directly generate the copy
             adata = self._adata_ref._getitem_copy((self._oidx, self._vidx))
@@ -965,11 +991,12 @@ class AnnData(IndexMixin):
 
     @property
     def obsm(self):
-        """Multi-dimensional annotation of observations (mutable structured `np.ndarray`).
+        """Multi-dimensional annotation of observations
+        (mutable structured :class:`np.ndarray`).
 
-        Stores for each key, a two or higher-dimensional `np.ndarray` of length
-        `n_obs`. Is sliced with `data` and `obs` but behaves otherwise like a
-        `dict`.
+        Stores for each key, a two or higher-dimensional :class:`np.ndarray` of length
+        ``n_obs``. Is sliced with ``data`` and ``obs`` but behaves otherwise like a
+        :class:`dict`.
         """
         return self._obsm
 
@@ -987,11 +1014,12 @@ class AnnData(IndexMixin):
 
     @property
     def varm(self):
-        """Multi-dimensional annotation of variables/ features (mutable structured `np.ndarray`).
+        """Multi-dimensional annotation of variables/ features
+        (mutable structured :class:`np.ndarray`).
 
-        Stores for each key, a two or higher-dimensional `np.ndarray` of length
-        `n_vars`. Is sliced with `data` and `var` but behaves otherwise like a
-        `dict`.
+        Stores for each key, a two or higher-dimensional :class:`np.ndarray` of length
+        ``n_vars``. Is sliced with ``data`` and ``var`` but behaves otherwise like a
+        :class:`dict`.
         """
         return self._varm
 
@@ -1030,19 +1058,19 @@ class AnnData(IndexMixin):
             utils.warn_names_duplicates('var', self._var)
 
     def obs_keys(self):
-        """List keys of observation annotation `.obs`."""
+        """List keys of observation annotation ``.obs``."""
         return self._obs.keys().tolist()
 
     def var_keys(self):
-        """List keys of variable annotation `var`."""
+        """List keys of variable annotation ``var``."""
         return self._var.keys().tolist()
 
     def obsm_keys(self):
-        """List keys of observation annotation `obsm`."""
+        """List keys of observation annotation ``obsm``."""
         return list(self._obsm.keys())
 
     def varm_keys(self):
-        """List keys of variable annotation `varm`."""
+        """List keys of variable annotation ``varm``."""
         return list(self._varm.keys())
 
     def uns_keys(self):
@@ -1051,22 +1079,22 @@ class AnnData(IndexMixin):
 
     @property
     def isbacked(self):
-        """`True` if object is backed on disk, `False` otherwise."""
+        """``True`` if object is backed on disk, ``False`` otherwise."""
         return self.filename is not None
 
     @property
     def isview(self):
-        """`True` if object is view of another AnnData object, `False` otherwise."""
+        """``True`` if object is view of another AnnData object, ``False`` otherwise."""
         return self._isview
 
     @property
     def filename(self):
-        """Change to backing mode by setting the filename of a `.h5ad` file.
+        """Change to backing mode by setting the filename of a ``.h5ad`` file.
 
         - Setting the filename writes the stored data to disk.
         - Setting the filename when the filename was previously another name
           moves the backing file from the previous file to the new file. If you
-          want to copy the previous file, use copy(filename='new_filename').
+          want to copy the previous file, use ``copy(filename='new_filename')``.
         """
         return self.file.filename
 
@@ -1107,7 +1135,7 @@ class AnnData(IndexMixin):
             del self.file[attr]
             self.file._file.create_dataset(attr, data=value)
 
-    def _normalize_indices(self, index):
+    def _normalize_indices(self, index: Union[Tuple, pd.Series, np.ndarray, slice, int, str, None]):
         # deal with tuples of length 1
         if isinstance(index, tuple) and len(index) == 1:
             index = index[0]
@@ -1125,7 +1153,7 @@ class AnnData(IndexMixin):
                 index = index[0].values, index[1]
             # one of the two has to be a slice
             if not (isinstance(index[0], slice) or isinstance(index[1], slice)):
-                if (isinstance(index[0], (int, str, None)) and isinstance(index[1], (int, str, None))):
+                if isinstance(index[0], (int, str, None)) and isinstance(index[1], (int, str, None)):
                     pass  # two scalars are fine
                 else:
                     raise ValueError(
@@ -1226,7 +1254,7 @@ class AnnData(IndexMixin):
     def _inplace_subset_var(self, index):
         """Inplace subsetting along variables dimension.
 
-        Same as adata = adata[:, index], but inplace.
+        Same as ``adata = adata[:, index]``, but inplace.
         """
         if not self.isbacked:
             self._X = self._X[:, index]
@@ -1244,7 +1272,7 @@ class AnnData(IndexMixin):
     def _inplace_subset_obs(self, index):
         """Inplace subsetting along variables dimension.
 
-        Same as adata = adata[index, :], but inplace.
+        Same as ``adata = adata[index, :]``, but inplace.
         """
         if not self.isbacked:
             self._X = self._X[index, :]
@@ -1273,7 +1301,7 @@ class AnnData(IndexMixin):
 
     def _get_var_array(self, k):
         """Get an array along the variables dimension by first looking up
-        var_keys and then obs_names."""
+        ``var_keys`` and then ``obs_names``."""
         x = (self._var[k] if k in self.var_keys()
              else self[k] if k in set(self.obs_names)
              else None)
@@ -1316,7 +1344,7 @@ class AnnData(IndexMixin):
     def copy(self, filename=None):
         """Full copy, optionally on disk."""
         if not self.isbacked:
-            return AnnData(self._X.copy(),
+            return AnnData(self._X.copy() if self._X is not None else None,
                            self._obs.copy(),
                            self._var.copy(),
                            # deepcopy on DictView does not work and is unnecessary
@@ -1336,44 +1364,49 @@ class AnnData(IndexMixin):
                 copyfile(self.filename, filename)
             return AnnData(filename=filename)
 
-    def concatenate(self, *adatas, join='inner', batch_key='batch',
-                    batch_categories=None, index_unique='-'):
+    def concatenate(
+        self, *adatas: 'AnnData',
+        join: str = 'inner',
+        batch_key: str = 'batch',
+        batch_categories: Sequence[Any] = None,
+        index_unique: Optional[str] = '-'
+    ) -> 'AnnData':
         """Concatenate along the observations axis.
 
-        The `.uns`, `.varm` and `.obsm` attributes are ignored.
+        The ``.uns``, ``.varm`` and ``.obsm`` attributes are ignored.
 
-        Currently, this works only in 'memory' mode.
+        Currently, this works only in ``'memory'`` mode.
 
         Parameters
         ----------
-        adatas : :class:`~anndata.AnnData`
+        adatas
             AnnData matrices to concatenate with. Each matrix is referred to as
-            a 'batch'.
-        join: `str` (default: 'inner')
+            a “batch”.
+        join
             Use intersection (``'inner'``) or union (``'outer'``) of variables.
-        batch_key : `str` (default: 'batch')
-            Add the batch annotation to `.obs` using this key.
-        batch_categories : list, optional (default: `range(len(adatas)+1)`)
-            Use these as categories for the batch annotation.
-        index_unique : {`None`, '-', str}, optional (default: `None`)
+        batch_key
+            Add the batch annotation to ``.obs`` using this key.
+        batch_categories
+            Use these as categories for the batch annotation. By default, use increasing numbers.
+        index_unique
             Make the index unique by joining the existing index names with the
-            batch category, using `index_unique`: '-', for instance. Provide
-            `None` to keep existing indices.
+            batch category, using ``index_unique='-'``, for instance. Provide
+            ``None`` to keep existing indices.
 
         Returns
         -------
-        adata : :class:`~anndata.AnnData`
-            The concatenated AnnData, where `adata.obs[batch_key]` stores a
-            categorical variable labeling the batch.
+        :class:`~anndata.AnnData`
+            The concatenated :class:`~anndata.AnnData`, where ``adata.obs[batch_key]``
+            stores a categorical variable labeling the batch.
 
         Notes
         -----
 
         .. warning::
 
-            If you use `join='outer'` this fills 0s for sparse data when
-            variables are absent in a batch. Use this with care. Dense data is
-            filled with NaNs. See the examples.
+           If you use ``join='outer'`` this fills 0s for sparse data when
+           variables are absent in a batch. Use this with care. Dense data is
+           filled with ``NaN``s. See the examples.
 
         Examples
         --------
@@ -1525,7 +1558,7 @@ class AnnData(IndexMixin):
                [0., 6., 5., 0.]], dtype=float32)
         """
         if self.isbacked:
-            raise ValueErro(
+            raise ValueError(
                 'Currently, concatenate does only work in \'memory\' mode.')
 
         if len(adatas) == 0:
@@ -1548,6 +1581,8 @@ class AnnData(IndexMixin):
         mergers = dict(inner=set.intersection, outer=set.union)
         var_names_reduce = reduce(mergers[join], (set(ad.var_names) for ad in all_adatas))
         # restore order of initial var_names, append non-sortable names at the end
+        # see how this was done in the repo at commit state
+        # 40a24f
         var_names = []
         for v in all_adatas[0].var_names:
             if v in var_names_reduce:
@@ -1565,22 +1600,34 @@ class AnnData(IndexMixin):
         out_shape = (sum(a.n_obs for a in all_adatas), len(var_names))
 
         any_sparse = any(issparse(a.X) for a in all_adatas)
-        if any_sparse:
-            X = sparse.lil_matrix(out_shape, dtype=self.X.dtype)
+        if join == 'outer':
+            if any_sparse:
+                X = sparse.lil_matrix(out_shape, dtype=self.X.dtype)
+            else:
+                X = np.empty(out_shape, dtype=self.X.dtype)
+                X[:] = np.nan
         else:
-            X = np.empty(out_shape, dtype=self.X.dtype)
-            X[:] = np.nan
+            Xs = []
 
         var = pd.DataFrame(index=var_names)
 
         obs_i = 0  # start of next adata’s observations in X
         out_obss = []
         for i, ad in enumerate(all_adatas):
-            vars_intersect = [v for v in var_names if v in ad.var_names]
+            if join == 'outer':
+                # only those names that are actually present in the current AnnData
+                vars_intersect = [v for v in var_names if v in ad.var_names]
+            else:
+                vars_intersect = var_names
 
             # X
-            X[obs_i:obs_i+ad.n_obs,
-              var_names.isin(vars_intersect)] = ad[:, vars_intersect].X
+            if join == 'outer':
+                # this is pretty slow, I guess sparse matrices shouldn't be
+                # constructed like that
+                X[obs_i:obs_i+ad.n_obs,
+                  var_names.isin(vars_intersect)] = ad[:, vars_intersect].X
+            else:
+                Xs.append(ad[:, vars_intersect].X)
             obs_i += ad.n_obs
 
             # obs
@@ -1600,8 +1647,13 @@ class AnnData(IndexMixin):
                 new_c = c + (index_unique if index_unique is not None else '-') + categories[i]
                 var.loc[vars_intersect, new_c] = ad.var.loc[vars_intersect, c]
 
-        if any_sparse:
-            X = X.tocsr()
+        if join == 'inner':
+            if any_sparse:
+                sparse_format = all_adatas[0].X.getformat()
+                from scipy.sparse import vstack
+                X = vstack(Xs).asformat(sparse_format)
+            else:
+                X = np.concatenate(Xs)
 
         obs = pd.concat(out_obss)
 
@@ -1653,17 +1705,22 @@ class AnnData(IndexMixin):
                              'columns of `X` ({}), but has {} rows.'
                              .format(self._n_vars, len(self._varm)))
 
-    def write(self, filename=None, compression='gzip', compression_opts=None):
+    def write(
+        self,
+        filename: Optional[Union[Path, str]] = None,
+        compression: Optional[str] = 'gzip',
+        compression_opts: Union[int, Any] = None,
+    ):
         """Write `.h5ad`-formatted hdf5 file and close a potential backing file.
 
         Parameters
         ----------
-        filename : `str`, optional (default: ``.filename``)
+        filename
             Filename of data file. Defaults to backing file.
-        compression : `None` or {'gzip', 'lzf'}, optional (default: `'gzip'`)
-            See http://docs.h5py.org/en/latest/high/dataset.html.
-        compression_opts : `int`, optional (default: `None`)
-            See http://docs.h5py.org/en/latest/high/dataset.html.
+        compression : ``None`` or {``'gzip'``, ``'lzf'``}
+            See the :ref:`h5py filter pipeline <h5py:dataset_compression>`.
+        compression_opts
+            See the :ref:`h5py filter pipeline <h5py:dataset_compression>`.
         """
         from .readwrite.write import _write_h5ad
 
@@ -1675,7 +1732,7 @@ class AnnData(IndexMixin):
         if self.isbacked:
             self.file.close()
 
-    def write_csvs(self, dirname, skip_data=True, sep=','):
+    def write_csvs(self, dirname: Union[Path, str], skip_data: bool = True, sep: str = ','):
         """Write annotation to `.csv` files.
 
         It is not possible to recover the full :class:`~anndata.AnnData` from the
@@ -1683,22 +1740,22 @@ class AnnData(IndexMixin):
 
         Parameters
         ----------
-        dirname : `str`
+        dirname
             Name of directory to which to export.
-        skip_data : `bool`, optional (default: True)
+        skip_data
              Skip the data matrix `.X`.
-        sep : `str`, optional (default: ',')
+        sep
              Separator for the data.
         """
         from .readwrite.write import write_csvs
         write_csvs(dirname, self, skip_data=skip_data, sep=sep)
 
-    def write_loom(self, filename):
+    def write_loom(self, filename: Union[Path, str]):
         """Write `.loom`-formatted hdf5 file.
 
         Parameters
         ----------
-        filename : `str`
+        filename
             The filename.
         """
         from .readwrite.write import write_loom
@@ -1846,70 +1903,66 @@ class AnnData(IndexMixin):
     # --------------------------------------------------------------------------
 
     @property
+    @utils.deprecated('X')
     def data(self):
-        """Deprecated access to X."""
-        print('DEPRECATION WARNING: use attribute `.X` instead of `.data`, '
-              '`.data` will be removed in the future.')
         return self.X
 
     @data.setter
+    @utils.deprecated('X')
     def data(self, value):
-        print('DEPRECATION WARNING: use attribute `.X` instead of `.data`, '
-              '`.data` will be removed in the future.')
         self.X = value
 
     @property
+    @utils.deprecated('n_obs')
     def n_smps(self):
-        """Deprecated access to `.n_obs`."""
         return self.n_obs
 
     @property
+    @utils.deprecated('obs')
     def smp(self):
-        """Deprecated access to `.obs`."""
         return self.obs
 
-    # for backwards compat
     @smp.setter
+    @utils.deprecated('obs')
     def smp(self, value):
         self.obs = value
 
     @property
+    @utils.deprecated('uns')
     def add(self):
-        """Deprecated access to `.uns`, remains for backwards compatibility."""
-        # FutureWarning and DeprecationWarning are not visible by default, use print
-        print('DEPRECATION WARNING: use attribute `.uns` instead of `.add`, '
-              '`.add` will be removed in the future.')
         return self.uns
 
     @add.setter
+    @utils.deprecated('uns')
     def add(self, value):
-        print('DEPRECATION WARNING: use attribute `.uns` instead of `.add`, '
-              '`.add` will be removed in the future.')
         self.uns = value
 
     @property
+    @utils.deprecated('obsm')
     def smpm(self):
-        """Multi-dimensional annotation of observations (mutable structured `np.ndarray`)        """
         return self.obsm
 
     @smpm.setter
+    @utils.deprecated('obsm')
     def smpm(self, value):
         self.obsm = value
 
     @property
+    @utils.deprecated('obs_names')
     def smp_names(self):
         return self.obs_names
 
     @smp_names.setter
+    @utils.deprecated('obs_names')
     def smp_names(self, names):
         self.obs_names = names
 
+    @utils.deprecated('obs_keys')
     def smp_keys(self):
-        """List keys of observation annotation `obs`."""
         return self.obs_keys()
 
+    @utils.deprecated('obsm_keys')
     def smpm_keys(self):
-        """List keys of observation annotation `obsm`."""
         return self.obsm_keys()
 
     def _clean_up_old_format(self, uns):
@@ -1959,3 +2012,16 @@ class AnnData(IndexMixin):
         values = getattr(self, a)[keys].values
         getattr(self, a).drop(keys, axis=1, inplace=True)
         return values
+
+    def chunked_X(self, chunk_size=None):
+        if chunk_size is None:
+            # Should be some adaptive code
+            chunk_size = 6000
+        start = 0
+        n = self.n_obs
+        for _ in range(int(n // chunk_size)):
+            end = start + chunk_size
+            yield (self.X[start:end], start, end)
+            start = end
+        if start < n:
+            yield (self.X[start:n], start, n)
